import os, pathlib
import shutil
import glob
import pickle

from argparse import ArgumentParser
import SimpleITK as sitk

from imgtools.ops import StructureSetToSegmentation, ImageAutoInput, ImageAutoOutput, Resample
from imgtools.pipeline import Pipeline
from joblib import Parallel, delayed

###############################################################
# Example usage:
# python radcure_simple.py ./data/RADCURE/data ./RADCURE_output
###############################################################


class AutoPipeline(Pipeline):
    """Example processing pipeline for the RADCURE dataset.
    This pipeline loads the CT images and structure sets, re-samples the images,
    and draws the GTV contour using the resampled image.
    """

    def __init__(self,
                 input_directory,
                 output_directory,
                 modalities="CT",
                 spacing=(1., 1., 0.),
                 n_jobs=-1,
                 visualize=False,
                 missing_strategy="drop",
                 show_progress=False,
                 warn_on_error=False):

        super().__init__(
            n_jobs=n_jobs,
            missing_strategy=missing_strategy,
            show_progress=show_progress,
            warn_on_error=warn_on_error)

        # pipeline configuration
        self.input_directory = input_directory
        self.output_directory = output_directory
        self.spacing = spacing
        self.existing = [None] #self.existing_patients()

        #input operations
        self.input = ImageAutoInput(input_directory, modalities, n_jobs, visualize)
        
        self.output_df_path = os.path.join(self.output_directory, "dataset.csv")
        #Output component table
        self.output_df = self.input.df_combined
        #Name of the important columns which needs to be saved    
        self.output_streams = self.input.output_streams
        
        # image processing ops
        self.resample = Resample(spacing=self.spacing)
        self.make_binary_mask = StructureSetToSegmentation(roi_names=[], continuous=False)

        # output ops
        self.output = ImageAutoOutput(self.output_directory, self.output_streams)

        #Make a directory
        if not os.path.exists(os.path.join(self.output_directory,".temp")):
            os.mkdir(os.path.join(self.output_directory,".temp"))


    def process_one_subject(self, subject_id):
        """Define the processing operations for one subject.
        This method must be defined for all pipelines. It is used to define
        the preprocessing steps for a single subject (note: that might mean
        multiple images, structures, etc.). During pipeline execution, this
        method will receive one argument, subject_id, which can be used to
        retrieve inputs and save outputs.
        Parameters
        ----------
        subject_id : str
           The ID of subject to process
        """
        #Check if the subject_id has already been processed
        if os.path.exists(os.path.join(self.output_directory,".temp",f'temp_{subject_id}.pkl')):
            print(f"{subject_id} already processed")
            return 

        print("Processing:", subject_id)

        read_results = self.input(subject_id)
        print(read_results)

        print(subject_id, " start")
        
        metadata = {}
        for i, colname in enumerate(self.output_streams):
            modality = colname.split("_")[0]

            # Taking modality pairs if it exists till _{num}
            output_stream = ("_").join([item for item in colname.split("_") if item.isnumeric()==False])

            # If there are multiple connections existing, multiple connections means two modalities connected to one modality. They end with _1
            mult_conn = colname.split("_")[-1].isnumeric()
            num = colname.split("_")[-1]

            print(output_stream)

            if read_results[i] is None:
                print("The subject id: {} has no {}".format(subject_id, colname))
                pass
            elif modality == "CT" or modality == 'MR':
                image = read_results[i]
                if len(image.GetSize()) == 4:
                    assert image.GetSize()[-1] == 1, f"There is more than one volume in this CT file for {subject_id}."
                    extractor = sitk.ExtractImageFilter()
                    extractor.SetSize([*image.GetSize()[:3], 0])
                    extractor.SetIndex([0, 0, 0, 0])    
                    
                    image = extractor.Execute(image)
                    print(image.GetSize())
                image = self.resample(image)
                #Saving the output
                self.output(subject_id, image, output_stream)
                metadata[f"size_{output_stream}"] = str(image.GetSize())
                print(subject_id, " SAVED IMAGE")
            elif modality == "RTDOSE":
                try: #For cases with no image present
                    doses = read_results[i].resample_dose(image)
                except:
                    Warning("No CT image present. Returning dose image without resampling")
                    doses = read_results[i]
                
                # save output
                if not mult_conn:
                    self.output(subject_id, doses, output_stream)
                else:
                    self.output(f"{subject_id}_{num}", doses, output_stream)
                metadata[f"size_{output_stream}"] = str(doses.GetSize())
                metadata[f"metadata_{colname}"] = [read_results[i].get_metadata()]
                print(subject_id, " SAVED DOSE")
            elif modality == "RTSTRUCT":
                #For RTSTRUCT, you need image or PT
                structure_set = read_results[i]
                conn_to = output_stream.split("_")[-1]

                # make_binary_mask relative to ct/pet
                if conn_to == "CT" or conn_to == "MR":
                    mask = self.make_binary_mask(structure_set, image)
                elif conn_to == "PT":
                    mask = self.make_binary_mask(structure_set, pet)
                else:
                    raise ValueError("You need to pass a reference CT or PT/PET image to map contours to.")
                
                # save output
                if not mult_conn:
                    self.output(subject_id, mask, output_stream)
                else:
                    self.output(f"{subject_id}_{num}", mask, output_stream)
                metadata[f"metadata_{colname}"] = [structure_set.roi_names]

                print(subject_id, "SAVED MASK ON", conn_to)
            elif modality == "PT":
                try:
                    #For cases with no image present
                    pet = read_results[i].resample_pet(image)
                except:
                    Warning("No CT image present. Returning PT/PET image without resampling.")
                    pet = read_results[i]

                if not mult_conn:
                    self.output(subject_id, pet, output_stream)
                else:
                    self.output(f"{subject_id}_{num}", pet, output_stream)
                metadata[f"size_{output_stream}"] = str(pet.GetSize())
                metadata[f"metadata_{colname}"] = [read_results[i].get_metadata()]
                print(subject_id, " SAVED PET")
        #Saving all the metadata in multiple text files
        with open(os.path.join(self.output_directory,".temp",f'{subject_id}.pkl'),'wb') as f:
            pickle.dump(metadata,f)
        return 
    
    def save_data(self):
        files = glob.glob(os.path.join(self.output_directory, ".temp", "*.pkl"))
        for file in files:
            filename = pathlib.Path(file).name
            subject_id = os.path.splitext(filename)[0]
            with open(file,"rb") as f:
                metadata = pickle.load(f)
            self.output_df.loc[subject_id, list(metadata.keys())] = list(metadata.values())
        self.output_df.to_csv(self.output_df_path)
        shutil.rmtree(os.path.join(self.output_directory, ".temp"))

    def run(self):
        """Execute the pipeline, possibly in parallel.
        """
        # Joblib prints progress to stdout if verbose > 50
        verbose = 51 if self.show_progress else 0

        subject_ids = self._get_loader_subject_ids()
        # Note that returning any SimpleITK object in process_one_subject is
        # not supported yet, since they cannot be pickled
        if os.path.exists(self.output_df_path):
            print("Dataset already processed...")
        else:
            Parallel(n_jobs=self.n_jobs, verbose=verbose)(
                    delayed(self._process_wrapper)(subject_id) for subject_id in subject_ids)
            self.save_data()
        

if __name__ == "__main__":
    parser = ArgumentParser("imgtools Automatic Processing Pipeline.")

    #arguments
    parser.add_argument("input_directory", type=str,
                        help="Path to top-level directory of dataset.")

    parser.add_argument("output_directory", type=str,
                        help="Path to output directory to save processed images.")

    parser.add_argument("--modalities", type=str, default="CT",
                        help="List of desired modalities. Type as string for ex: RTSTRUCT,CT,RTDOSE")
    
    parser.add_argument("--visualize", type=bool, default=False,
                        help="Whether to visualize the data graph")
    
    parser.add_argument("--spacing", nargs=3, type=float, default=(1., 1., 0.),
                        help="The resampled voxel spacing in  (x, y, z) directions.")

    parser.add_argument("--n_jobs", type=int, default=-1,
                        help="The number of parallel processes to use.")

    parser.add_argument("--show_progress", action="store_true",
                        help="Whether to print progress to standard output.")

    args = parser.parse_args()
    pipeline = AutoPipeline(args.input_directory,
                            args.output_directory,
                            modalities=args.modalities,
                            spacing=args.spacing,
                            n_jobs=args.n_jobs,
                            visualize=args.visualize,
                            show_progress=args.show_progress)

    print(f'starting Pipeline...')
    pipeline.run()


    print(f'finished Pipeline!')


import os, pathlib
import glob
import re
from typing import Optional, List
from collections import namedtuple
import json

import pandas as pd
import SimpleITK as sitk
import nrrd
from pydicom import dcmread

from joblib import Parallel, delayed
from tqdm.auto import tqdm

from ..modules import StructureSet
from ..modules import Dose
from ..modules import PET
from ..utils.crawl import *



def read_image(path):
    return sitk.ReadImage(path)

def read_header(path):
    return nrrd.read_header(path)

def read_dicom_series(path: str,
                      series_id: Optional[str] = None,
                      recursive: bool = False) -> sitk.Image:
    """Read DICOM series as SimpleITK Image.

    Parameters
    ----------
    path
       Path to directory containing the DICOM series.

    recursive, optional
       Whether to recursively parse the input directory when searching for
       DICOM series,

    series_id, optional
       Specifies the DICOM series to load if multiple series are present in
       the directory. If None and multiple series are present, loads the first
       series found.


    Returns
    -------
    The loaded image.

    """
    reader = sitk.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(path,
                                                seriesID=series_id if series_id else "",
                                                recursive=recursive)
    reader.SetFileNames(dicom_names)
    
    # Configure the reader to load all of the DICOM tags (public+private):
    # By default tags are not loaded (saves time).
    # By default if tags are loaded, the private tags are not loaded.
    # We explicitly configure the reader to load tags, including the
    # private ones.
    reader.MetaDataDictionaryArrayUpdateOn()
    reader.LoadPrivateTagsOn()

    return reader.Execute()


def read_dicom_rtstruct(path):
    return StructureSet.from_dicom_rtstruct(path)

def read_dicom_rtdose(path):
    return Dose.from_dicom_rtdose(path)

def read_dicom_pet(path,series=None):
    return PET.from_dicom_pet(path,series, "SUV")

def read_dicom_auto(path, series=None):
    if path is None:
        return None
    dcms = glob.glob(os.path.join(path, "*.dcm"))
    meta = dcmread(dcms[0])
    modality = meta.Modality
    if modality == 'CT' or modality == 'MR':
        return read_dicom_series(path,series)
    elif modality == 'PT':
        return read_dicom_pet(path,series)
    # elif len(dcms) == 1:
    #     meta = dcmread(dcms[0])
    #     modality = meta.Modality
    elif modality == 'RTSTRUCT':
        return read_dicom_rtstruct(dcms[0])
    elif modality == 'RTDOSE':
        return read_dicom_rtdose(path)
    else:
        if len(dcms)==1:
            raise NotImplementedError
        else:
            print("There were no dicoms in this path.")
            return None

def read_segmentation(path):
    # TODO read seg.nrrd
    pass

class BaseLoader:
    def __getitem__(self, subject_id):
        raise NotImplementedError

    def __len__(self):
        return len(self.keys())

    def keys(self):
        raise NotImplementedError

    def items(self):
        return ((k, self[k]) for k in self.keys())

    def values(self):
        return (self[k] for k in self.keys())

    def get(self, subject_id, default=None):
        try:
            return self[subject_id]
        except KeyError:
            return default

class ImageCSVLoader(BaseLoader):
    def __init__(self,
                 csv_path_or_dataframe,
                 colnames=[],
                 seriesnames=[],
                 id_column=None,
                 expand_paths=False,
                 readers=[read_image]):

        self.expand_paths = expand_paths
        self.readers = readers

        self.colnames = colnames
        self.seriesnames = seriesnames

        if isinstance(csv_path_or_dataframe, str):
            if id_column is not None and id_column not in colnames:
                colnames.append(id_column)
            self.paths = pd.read_csv(csv_path_or_dataframe,
                                     index_col=id_column)
        elif isinstance(csv_path_or_dataframe, pd.DataFrame):
            self.paths = csv_path_or_dataframe
            if id_column:
                self.paths = self.paths.set_index(id_column)
            if not self.colnames:
                self.colnames = self.paths.columns
        else:
            raise ValueError(f"Expected a path to csv file or pd.DataFrame, not {type(csv_path_or_dataframe)}.")

        if not isinstance(readers, list):
            readers = [readers] * len(self.colnames)

        self.output_tuple = namedtuple("Output", self.colnames)

    def __getitem__(self, subject_id):
        row = self.paths.loc[subject_id]
        paths = {col: row[col] for col in self.colnames}
        series = {col: row[col] for col in self.seriesnames}
        if self.expand_paths:
            # paths = {col: glob.glob(path)[0] for col, path in paths.items()}
            paths = {col: glob.glob(path)[0] if pd.notna(path) else None for col, path in paths.items()}
        outputs = {col: self.readers[i](path,series["series_"+("_").join(col.split("_")[1:])]) for i, (col, path) in enumerate(paths.items())}
        return self.output_tuple(**outputs)

    def keys(self):
        return list(self.paths.index)

    def items(self):
        return ((k, self[k]) for k in self.keys())


class ImageFileLoader(BaseLoader):
    def __init__(self,
                 root_directory,
                 get_subject_id_from="filename",
                 subdir_path=None,
                 exclude_paths=[],
                 reader=read_image):

        self.root_directory = root_directory
        self.get_subject_id_from = get_subject_id_from
        self.subdir_path = subdir_path
        self.exclude_paths = []
        for path in exclude_paths:
            if not path.startswith(self.root_directory):
                full_paths = glob.glob(os.path.join(root_directory, path))
                self.exclude_paths.extend(full_paths)
            else:
                full_path = path
                self.exclude_paths.append(full_path)
        self.reader = reader

        self.paths = self._generate_paths()

    def _generate_paths(self):
        paths = {}
        for f in os.scandir(self.root_directory):
            if f.path in self.exclude_paths:
                continue
            subject_dir_path = f.path
            if self.subdir_path:
                full_path = os.path.join(subject_dir_path, self.subdir_path)
            else:
                full_path = subject_dir_path
            try:
                full_path = glob.glob(full_path)[0]
            except IndexError:
                continue
            if os.path.isdir(full_path):
                full_path = os.path.join(full_path, "")
            subject_dir_name = os.path.basename(os.path.normpath(subject_dir_path))
            subject_id = self._extract_subject_id_from_path(full_path, subject_dir_name)
            paths[subject_id] = full_path
        return paths

    def _extract_subject_id_from_path(self, full_path, subject_dir_name):
        filename, _ = os.path.splitext(os.path.basename(full_path))
        if isinstance(self.get_subject_id_from, str):
            if self.get_subject_id_from == "filename":
                subject_id = filename
            elif self.get_subject_id_from == "subject_directory":
                subject_id = subject_dir_name
            else:
                subject_id = re.search(self.get_subject_id_from, full_path)[0]
        else:
            return self.get_subject_id_from(full_path, filename, subject_dir_name)
        return subject_id

    def __getitem__(self, subject_id):
        path = self.paths[subject_id]
        return self.reader(path)

    def keys(self):
        return self.paths.keys()



# class CombinedLoader(BaseLoader):
#     def __init__(self, **kwargs):
#         self.loaders = kwargs
#         self.output_tuple = namedtuple("Output", list(self.loaders.keys()))

#     def __getitem__(self, subject_id):
#         outputs = {name: loader[subject_id] for name, loader in self.loaders.items()}
#         return self.output_tuple(**outputs)

#     def keys(self):
#         return set(chain.from_iterable(loader.keys() for loader in self.loaders))

#     def items(self):
#         return ((k, self[k]) for k in self.keys())



import re
from warnings import warn
from typing import Dict, List, Optional, Union

import numpy as np
import SimpleITK as sitk
from pydicom import dcmread
from itertools import groupby
from skimage.draw import polygon2mask

from .segmentation import Segmentation
from ..utils import physical_points_to_idxs


def _get_roi_points(rtstruct, roi_index):
    return [np.array(slc.ContourData).reshape(-1, 3) for slc in rtstruct.ROIContourSequence[roi_index].ContourSequence]


class StructureSet:
    def __init__(self, roi_points: Dict[str, np.ndarray]):
        self.roi_points = roi_points

    @classmethod
    def from_dicom_rtstruct(cls, rtstruct_path: str) -> 'StructureSet':
        rtstruct = dcmread(rtstruct_path, force=True)
        roi_names = [roi.ROIName for roi in rtstruct.StructureSetROISequence]

        roi_points = {}
        for i, name in enumerate(roi_names):
            try:
                roi_points[name] = _get_roi_points(rtstruct, i)
            except AttributeError:
                warn(f"Could not get points for ROI {name} (in {rtstruct_path}).")
        return cls(roi_points)

    @property
    def roi_names(self) -> List[str]:
        return list(self.roi_points.keys())

    def _assign_labels(self, names, force_missing=False):
        """
        args
        ----
        force_missing
            What does force_missing do?
        """
        labels = {}
        cur_label = 0
        if names == self.roi_names:
            for i, name in enumerate(self.roi_names):
                labels[name] = i
        else:
            for j, pat in enumerate(names):
                if sorted(names) == sorted(list(labels.keys())): #checks if all ROIs have already been processed.
                    break
                if isinstance(pat, str):
                    matched = False
                    for i, name in enumerate(self.roi_names):
                        if re.fullmatch(pat, name, flags=re.IGNORECASE):
                            labels[name] = cur_label
                            cur_label += 1
                            matched = True
                    if force_missing and not matched:
                        labels[pat] = cur_label
                        cur_label += 1
                else:
                    matched = False
                    for subpat in pat:
                        for name in self.roi_names:
                            if re.fullmatch(subpat, name, flags=re.IGNORECASE):
                                labels[name] = cur_label
                                matched = True
                    if force_missing and not matched:
                        key = '_'.join(pat)
                        labels[key] = cur_label
                    cur_label += 1
        return labels

    def to_segmentation(self, reference_image: sitk.Image,
                        roi_names: Optional[List[Union[str, List[str]]]] = None,
                        force_missing: bool = False,
                        continuous: bool = True) -> Segmentation:
        """Convert the structure set to a Segmentation object.

        Parameters
        ----------
        reference_image
            Image used as reference geometry.
        roi_names
            List of ROI names to export. Both full names and
            case-insensitive regular expressions are allowed.
            All labels within one sublist will be assigned
            the same label.
        force_missing
            If True, the number of labels in the output will
            be equal to `len(roi_names)`, with blank slices for
            any missing labels. Otherwise, missing ROI names
            will be excluded.

        Returns
        -------
        Segmentation
            The segmentation object.

        Notes
        -----
        If `roi_names` contains lists of strings, each matching
        name within a sublist will be assigned the same label. This means
        that `roi_names=['pat']` and `roi_names=[['pat']]` can lead
        to different label assignments, depending on how many ROI names
        match the pattern. E.g. if `self.roi_names = ['fooa', 'foob']`,
        passing `roi_names=['foo(a|b)']` will result in a segmentation with 
        two labels, but passing `roi_names=[['foo(a|b)']]` will result in
        one label for both `'fooa'` and `'foob'`.

        In general, the exact ordering of the returned labels cannot be
        guaranteed (unless all patterns in `roi_names` can only match
        a single name or are lists of strings).
        """
        if not roi_names:
            roi_names = self.roi_names
        if isinstance(roi_names, str):
            roi_names = [roi_names]
       
        labels = self._assign_labels(roi_names, force_missing)
        print("labels:", labels)
        if not labels:
            raise ValueError(f"No ROIs matching {roi_names} found in {self.roi_names}.")

        size = reference_image.GetSize()[::-1] + (max(labels.values()) + 1,)

        mask = np.zeros(size, dtype=np.uint8)

        for name, label in labels.items():
            physical_points = self.roi_points.get(name, np.array([]))
            if len(physical_points) == 0:
                continue # allow for missing labels, will return a blank slice

            mask_points = physical_points_to_idxs(reference_image, physical_points, continuous=continuous)

            for contour in mask_points:
                z, slice_points = np.unique(contour[:, 0]), contour[:, 1:]
                if len(z) == 1:
                    # assert len(z) == 1, f"This contour ({name}) spreads across more than 1 slice."
                    z = z[0]
                    slice_mask = polygon2mask(size[1:-1], slice_points)
                    mask[z, :, :, label] += slice_mask

        
        mask[mask > 1] = 1        
        mask = sitk.GetImageFromArray(mask, isVector=True)
        mask.CopyInformation(reference_image)
        seg_roi_names = {"_".join(k): v for v, k in groupby(labels, key=lambda x: labels[x])}
        mask = Segmentation(mask, roi_names=seg_roi_names)

        return mask

    def __repr__(self):
        return f"<StructureSet with ROIs: {self.roi_names!r}>"



from functools import wraps

import numpy as np
import SimpleITK as sitk

from ..utils import array_to_image, image_to_array


def accepts_segmentations(f):
    @wraps(f)
    def wrapper(img, *args, **kwargs):
         result = f(img, *args, **kwargs)
         if isinstance(img, Segmentation):
             result = sitk.Cast(result, sitk.sitkVectorUInt8)
             return Segmentation(result, roi_names=img.roi_names)
         else:
             return result
    return wrapper


def map_over_labels(segmentation, f, include_background=False, return_segmentation=True, **kwargs):
    if include_background:
        labels = range(segmentation.num_labels + 1)
    else:
        labels = range(1, segmentation.num_labels + 1)
    res = [f(segmentation.get_label(label=label), **kwargs) for label in labels]
    if return_segmentation and isinstance(res[0], sitk.Image):
        res = [sitk.Cast(r, sitk.sitkUInt8) for r in res]
        res = Segmentation(sitk.Compose(*res), roi_names=segmentation.roi_names)
    return res


class Segmentation(sitk.Image):
    def __init__(self, segmentation, roi_names=None):
        super().__init__(segmentation)
        self.num_labels = self.GetNumberOfComponentsPerPixel()
        if not roi_names:
            self.roi_names = {f"label_{i}": i for i in range(1, self.num_labels+1)}
        else:
            self.roi_names = roi_names
            if 0 in self.roi_names.values():
                self.roi_names = {k : v+1 for k, v in self.roi_names.items()}
        if len(self.roi_names) != self.num_labels:
            for i in range(1, self.num_labels+1):
                if i not in self.roi_names.values():
                    self.roi_names[f"label_{i}"] = i

    def get_label(self, label=None, name=None, relabel=False):
        if label is None and name is None:
            raise ValueError("Must pass either label or name.")

        if label is None:
            label = self.roi_names[name]

        if label == 0:
            # background is stored implicitly and needs to be computed
            label_arr = sitk.GetArrayViewFromImage(self)
            label_img = sitk.GetImageFromArray((label_arr.sum(-1) == 0).astype(np.uint8))
        else:
            label_img = sitk.VectorIndexSelectionCast(self, label - 1)
            if relabel:
                label_img *= label

        return label_img

    def to_label_image(self):
        arr, *_ = image_to_array(self)
        # TODO handle overlapping labels
        label_arr = np.where(arr.sum(-1) != 0, arr.argmax(-1) + 1, 0)
        label_img = array_to_image(label_arr, reference_image=self)
        return label_img

    # TODO also overload other operators (arithmetic, etc.)
    # with some sensible behaviour

    def __getitem__(self, idx):
        res = super().__getitem__(idx)
        if isinstance(res, sitk.Image):
            res = Segmentation(res, self.roi_names)
        return res

    def __repr__(self):
        return f"<Segmentation with ROIs: {self.roi_names!r}>"



from argparse import ArgumentParser
import os
import glob
import json

import pandas as pd
from pydicom import dcmread
from tqdm import tqdm

from joblib import Parallel, delayed

def crawl_one(folder):
    database = {}
    for path, _, _ in os.walk(folder):
        # find dicoms
        dicoms = glob.glob(os.path.join(path, "*.dcm"))

        # instance (slice) information
        for dcm in dicoms:
            try:
                meta = dcmread(dcm, force=True)
                patient  = str(meta.PatientID)
                study    = str(meta.StudyInstanceUID)
                series   = str(meta.SeriesInstanceUID)
                instance = str(meta.SOPInstanceUID)

                reference_ct, reference_rs, reference_pl = " ", " ", " "
                try: #RTSTRUCT
                    reference_ct = str(meta.ReferencedFrameOfReferenceSequence[0].RTReferencedStudySequence[0].RTReferencedSeriesSequence[0].SeriesInstanceUID)
                except: 
                    try: #RTDOSE
                        reference_rs = str(meta.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID)
                    except:
                        pass
                    try:
                        reference_ct = str(meta.ReferencedImageSequence[0].ReferencedSOPInstanceUID)
                    except:
                        pass
                    try:
                        reference_pl = str(meta.ReferencedRTPlanSequence[0].ReferencedSOPInstanceUID)
                    except:
                        pass
                
                try:
                    reference_frame = str(meta.FrameOfReferenceUID)
                except:
                    try:
                        reference_frame = str(meta.ReferencedFrameOfReferenceSequence[0].FrameOfReferenceUID)
                    except:
                        reference_frame = ""
        
                try:
                    study_description = str(meta.StudyDescription)
                except:
                    study_description = ""

                try:
                    series_description = str(meta.SeriesDescription)
                except:
                    series_description = ""

                if patient not in database:
                    database[patient] = {}
                if study not in database[patient]:
                    database[patient][study] = {'description': study_description}
                if series not in database[patient][study]:
                    database[patient][study][series] = {'instances': [],
                                                        'instance_uid': instance,
                                                        'modality': meta.Modality,
                                                        'description': series_description,
                                                        'reference_ct': reference_ct,
                                                        'reference_rs': reference_rs,
                                                        'reference_pl': reference_pl,
                                                        'reference_frame': reference_frame,
                                                        'folder': path}
                database[patient][study][series]['instances'].append(instance)
            except:
                pass
    
    return database

def to_df(database_dict):
    df = pd.DataFrame()
    for pat in database_dict:
        for study in database_dict[pat]:
            for series in database_dict[pat][study]:
                if series != 'description':
                    df = df.append({'patient_ID': pat,
                                    'study': study,
                                    'study_description': database_dict[pat][study]['description'],
                                    'series': series,
                                    'series_description': database_dict[pat][study][series]['description'],
                                    'modality': database_dict[pat][study][series]['modality'],
                                    'instances': len(database_dict[pat][study][series]['instances']),
                                    'instance_uid': database_dict[pat][study][series]['instance_uid'],
                                    'reference_ct': database_dict[pat][study][series]['reference_ct'],
                                    'reference_rs': database_dict[pat][study][series]['reference_rs'],
                                    'reference_pl': database_dict[pat][study][series]['reference_pl'],
                                    'reference_frame': database_dict[pat][study][series]['reference_frame'],
                                    'folder': database_dict[pat][study][series]['folder']}, ignore_index=True)
    return df

def crawl(top, 
          n_jobs: int = -1):
    database_list = []
    folders = glob.glob(os.path.join(top, "*"))
    
    database_list = Parallel(n_jobs=n_jobs)(delayed(crawl_one)(os.path.join(top, folder)) for folder in tqdm(folders))

    # convert list to dictionary
    database_dict = {}
    for db in database_list:
        for key in db:
            database_dict[key] = db[key]
    
    # save one level above imaging folders
    parent, dataset  = os.path.split(top)
    
    # save as json
    with open(os.path.join(parent, f'imgtools_{dataset}.json'), 'w') as f:
        json.dump(database_dict, f, indent=4)
    
    # save as dataframe
    df = to_df(database_dict)
    df_path = os.path.join(parent, f'imgtools_{dataset}.csv')
    df.to_csv(df_path)
    
    return database_dict

if __name__ == "__main__":
    parser = ArgumentParser("Dataset DICOM Crawler")
    parser.add_argument("directory",
                         type=str,
                         help="Top-level directory of the dataset.")
    parser.add_argument("--n_jobs",
                         type=int,
                         default=16,
                         help="Number of parallel processes for multiprocessing.")

    args = parser.parse_args()
    db = crawl(args.directory, n_jobs=args.n_jobs)
    print("# patients:", len(db))



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    reqs = fh.read()
    
setup(
    name="med-imagetools",
    version="0.2.3",
    author="Michal Kazmierski, Sejin Kim, Vishwesh Ramanathan, Benjamin Haibe-Kains",
    author_email="benjamin.haibe.kains@utoronto.ca",
    description="Transparent and reproducible image processing pipelines in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bhklab/med-imagetools",
    install_requires=reqs,
    packages=find_packages(),
    extras_require={
        'debug': ['pyvis'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 2 - Pre-Alpha"
    ],
    python_requires='>=3.6',
)
